/**
 * @file unity.h
 * The Library to support parsing unit strings
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define UNITY_INTERNAL 1
#include "unitylib.h"           /* functions in this module */
#include "unit-definitions.h"
#include "function-definitions.h"
#include "unity-constants.h"
#include "known-syntaxes.h"
#include "config.h"

// Declare functions defined elsewhere
int yyparse(void);              /* from yacc */
int lex_setup(const char* s);   /* from lex.lex */
char lex_unexpected_character(void);

void yyerror(const char *s);

static const char* find_prefix_and_base(const char* s, int* pfx);
static char* u_copy_string(const char* s);


// Functions defined in this module
// Functions u_* are 'internal' to the library. 
// Some of these are additionally static.

/*
 * u_new_unit and u_function_application allocate new Unit objects.
 * If the parse succeeds,
 * these are subsequently owned and managed by a UnitExpression.  If
 * it does not succeed, however, then these aren't managed by
 * anything.  Keep track of them, therefore in a list within
 * track_unit.  See unity_parse_string, which manages them after an
 * unsuccessful parse.
 */
static const Unit** unit_tracking = NULL;
static int unit_tracking_size = 0;
static int unit_tracking_live = 0;
// (unit_tracking != NULL || unit_tracking_live == 0) initially true
static void track_unit(const Unit* u) 
{
    if (u == NULL) {
        // all's well (presumably), so stop tracking these units
        // (we don't free them, because they're now owned by the
        // UnitExpression they're part of, so we simply forget about them)
        unit_tracking_live = 0;
    } else {
        if (unit_tracking_live >= unit_tracking_size) {
            if (unit_tracking != NULL) {
                unit_tracking_size *= 2;
                unit_tracking = (const Unit**)realloc((void*)unit_tracking, unit_tracking_size*sizeof(Unit*));
            } else {
                // first time
                unit_tracking_size = 8; /* initial size */
                unit_tracking = (const Unit**)malloc(unit_tracking_size*sizeof(Unit*));
            }
        }
        unit_tracking[unit_tracking_live++] = u;
    }
    assert(unit_tracking != NULL || unit_tracking_live == 0);
    return;
}
/* Free all of the tracked units */
static void track_unit_free(void)
{
    int i;
    assert(unit_tracking != NULL || unit_tracking_live == 0);
    for (i=0; i<unit_tracking_live; i++) {
        u_free_unit(unit_tracking[i]);
    }
    unit_tracking_live = 0;
    return;
}
/* Is the tracking system empty? (for assertions) */
#define TRACK_UNIT_EMPTY_P (unit_tracking_live == 0)

Unit* u_new_unit(const char* unit_string, const float power, int syntax)
{
    const UnitDef* def;
    Unit* rval;

    if ((rval = (Unit*)malloc(sizeof(Unit))) == NULL) {
        fprintf(stderr, "Unable to allocate new unit\n");
        exit(1);
    }
    track_unit(rval);
    
    rval->power = power;            /* easy */
    rval->type = simple_unit_type;
    rval->next = NULL;

    if ((def = unity_get_unit(unit_string, syntax)) != NULL) {
        rval->u.prefix_power = 0;
        rval->u.base_unit_def = def;
        rval->u.base_unit_string = NULL;
    } else {
        const char* base_unit;
        int prefix_power;

        if ((base_unit = find_prefix_and_base(unit_string, &prefix_power)) != NULL) {
            rval->u.prefix_power = prefix_power;
            if ((def = unity_get_unit(base_unit, syntax)) == NULL) {
                rval->u.base_unit_string = u_copy_string(base_unit);
                rval->u.base_unit_def = NULL;
            } else {
                rval->u.base_unit_string = NULL;
                rval->u.base_unit_def = def;
            }
        } else {
            rval->u.prefix_power = 0;
            rval->u.base_unit_string = u_copy_string(unit_string);
            rval->u.base_unit_def = NULL;
        }
    }

    // structure invariant: precisely one of base_unit_string and base_unit_def is non-NULL
    assert(   (rval->u.base_unit_string == NULL && rval->u.base_unit_def != NULL) 
           || (rval->u.base_unit_string != NULL && rval->u.base_unit_def == NULL));
    // not an invariant, but always true on creation
    assert(rval->next == NULL);

    return rval;
}

Unit* u_function_application(const char* functionName,
                             const Unit* unit_sequence,
                             int syntax)
{
    Unit* U;
    const FunctionDef* def;
    
    if ((U = (Unit*)malloc(sizeof(Unit))) == NULL) {
        fprintf(stderr, "Can't allocate new unit; bailing!\n");
        exit(1);
    }
    track_unit(U);

    // easy ones first
    U->type = function_application_type;
    U->power = 1;
    U->next = NULL;

    if ((def = unity_get_function(functionName, syntax)) == NULL) {
        U->f.definition = NULL;
        U->f.name = u_copy_string(functionName);
    } else {
        U->f.definition = def;
        U->f.name = NULL;
    }
    assert(   (U->f.name == NULL && U->f.definition != NULL)
           || (U->f.name != NULL && U->f.definition == NULL));
    U->f.operand = unit_sequence;

    // not an invariant, but always true on creation
    assert(U->next == NULL);

    return U;
}

/*
 * Frees a previously obtained unit.
 */
void u_free_unit(const Unit* u) 
{
    assert(u != NULL);
    switch(u->type) {
      case simple_unit_type:
        if (u->u.base_unit_string != NULL) {
            free((void*)u->u.base_unit_string);
        }
        break;

      case function_application_type:
        if (u->f.name != NULL) {
            free((void*)u->f.name);
        }
        break;

      default:
        assert(0);
    }
    free((void*)u);
}

/**
 * Frees a previously obtained unit-expression
 */
void unity_free_expression(const UnitExpression* ue) 
{
    const Unit* u;
    const Unit* next;

    assert(ue != NULL);
    for (u=ue->unit_sequence; u!=NULL; u=next) {
        next = u->next;
        u_free_unit(u);
    }
    free((void*)ue);
}

/* debugging function */
void uu_display_unit(const char* prefix, const Unit* p, FILE* out, const char* suffix)
{
    fputs(prefix, out);
    while (p != NULL) {
        switch (p->type) {
          case simple_unit_type:
            fprintf(out, " %d/%s/%g",
                    p->u.prefix_power,
                    (p->u.base_unit_string != NULL
                     ? p->u.base_unit_string
                     : p->u.base_unit_def->name),
                    p->power);
            break;

          case function_application_type:
            fprintf(out, " %s", p->f.definition->description);
            uu_display_unit("(", p->f.operand, out, ")");
            fprintf(out, "/%g", p->power);
            break;

          default:
            assert(0);
        }
        p = p->next;
    }
    fputs(suffix, out);
}

/*
 * Divide the numerator by the denominator.
 * Nothing fancy -- just flip the powers of the denominator and append it to the numerator.
 * If either of the arguments is NULL, return NULL.
 * Modifies both.
 */
Unit* u_divide_units(Unit* num, Unit* den)
{
    Unit* p;

    if (num == NULL || den == NULL) {
        return NULL;
    }
    for (p=den; p!=NULL; p=p->next) {
        p->power *= -1;
    }
    // find the end of the numerator string
    for (p=num; p->next != NULL; p=p->next) {
        // do nothing
    }
    p->next = den;
    return num;
}

Unit* u_unit_reciprocal(Unit* u)
{
    Unit* p;

    assert(u != NULL);
    for (p=u; p!=NULL; p=p->next) {
        p->power *= -1;
    }
    return u;
}

Unit* u_unit_append(Unit* u1, Unit* u2)
{
    Unit* p;

    assert(u1 != NULL && u2 != NULL);
    for (p=u1; p->next!=NULL; p=p->next) {
        // nothing
    }
    p->next = u2;
    return u1;
}

/*
 * Return a pointer to a copy of the given non-null string.
 * The pointer must be freed by the caller.
 */
static char* u_copy_string(const char* s) 
{
    int len;

    assert(s != NULL);          /* programming error if s is NULL */
    len = strlen(s);
    char* rval;
    if ((rval = (char*)malloc(len+1)) == NULL) {
        fprintf(stderr, "Can't malloc %d bytes for copy_string\n", len+1);
        exit(1);
    }
    memcpy(rval, s, len+1);
    return rval;
}

/*
 * Translate the character prefix to the base-10 log it corresponds to.
 * @return a non-zero integer.
 */
int u_prefix_to_power(const char pfx) 
{
    int rval = 0; /* initialising keeps the compiler happy, and allows the assertion check at the end */
    switch (pfx) {
      case 'y': rval = -24; break;
      case 'z': rval = -21; break;
      case 'a': rval = -18; break;
      case 'f': rval = -15; break;
      case 'p': rval = -12; break;
      case 'n': rval = -9; break;
      case 'u': rval = -6; break;
      case 'm': rval = -3; break;
      case 'c': rval = -2; break;
      case 'd': rval = -1; break;
      case UNITY_PREFIX_DECA:
        rval = 1; break; /* deca */
      case 'h': rval = 2; break;
      case 'k': rval = 3; break;
      case 'M': rval = 6; break;
      case 'G': rval = 9; break;
      case 'T': rval = 12; break;
      case 'P': rval = 15; break;
      case 'E': rval = 18; break;
      case 'Z': rval = 21; break;
      case 'Y': rval = 24; break;
      default:
        fprintf(stderr, "Impossible prefix: %c\n", pfx);
        assert(0);
        break;
    }
    assert(rval != 0);
    return rval;
}

typedef int (*parsef_ptr)(void);
int fits_yyparse(void);
int ogip_yyparse(void);
int cds_yyparse(void);
int vou_yyparse(void);
static struct {
    const char* name;
    parsef_ptr parser;
} parser_list[] = {
    { "fits", &fits_yyparse },
    { "ogip", &ogip_yyparse },
    { "cds",  &cds_yyparse },
    { "vou",  &vou_yyparse },
};
static int parser_list_size = sizeof(parser_list) / sizeof(parser_list[0]);

static parsef_ptr get_parser_for_syntax(int syntax) 
{
    const char* name = unity_get_syntax_name(syntax);
    // Sanity check (this would ideally be a preprocessor-time check, but that's not possible)
    assert(parser_list_size == U_NSYNTAXES);
    if (name != NULL) {
        int i;
        for (i=0; i<parser_list_size; i++) {
            if (strcmp(name, parser_list[i].name) == 0) {
                return parser_list[i].parser; /* JUMP OUT */
            }
        }
    }
    return NULL;
}

/**
 * Return the available parser names.  If this is impossible for some
 * reason (which can really only be memory exhaustion, which should
 * never happen), print an error message and return NULL.
 * The returned list is statically allocated, and should not be freed
 * by the caller.
 * @return a NULL-terminated list of pointers to parser names
 */
const char** unity_parser_names(void)
{
    static const char** res = NULL;
    if (res == NULL) {
        // initialise
        int i;
        if ((res = (const char**)malloc(sizeof(char*)*(parser_list_size+1))) == NULL) {
            fprintf(stderr, "Can't allocate %ld bytes for unity_parser_names\n", sizeof(char*)*(parser_list_size+1));
            exit(1);
            //return NULL;
        }
        assert(res != NULL);
        for (i=0; i<parser_list_size; i++) {
            res[i] = parser_list[i].name;
        }
        res[parser_list_size] = NULL; /* terminate the list */
    }
    return res;
}

/**
 * Obtain the parser ID from a parser name.
 * The recognised parser names are those returned by {@link #unity_parser_names},
 * and the returned IDs can be given as arguments to {@link #unity_parse_string}.
 *
 * <p>The numerical value of the returned parser identifier is one of
 * the constants <code>UNITY_SYNTAX_...</code>.  It's value is not
 * guaranteed, and may change between library releases.
 * @param parser_name a name such as "fits"
 * @return a non-negative integer; or negative if the name is not recognised
 */
int unity_identify_parser(const char* parser_name)
{
    int i;
    for (i=0; i<parser_list_size; i++) {
        if (strcmp(parser_name, parser_list[i].name) == 0) {
            return i;
        }
    }
    return -1;
}

/**
 * Finds a parser name from an ID
 * @param parser_id one of the constants <code>UNITY_SYNTAX_*</code>
 * @return a name for the parser, or NULL if the constant isn't recognised
 */
const char* unity_parser_name(const int parser_id)
{
    if (parser_id < 0 || parser_id >= U_NSYNTAXES) {
        return NULL;
    } else {
        return parser_list[parser_id].name;
    }
}

/* The parser reports the parse result by calling u_receive_result
   with the result of the parse.  This is reported to the caller by
   unity_parse_string, which then sets parse_result back to NULL.
   These two functions communicate via the parse_result variable,
   which should be NULL before when unity_parse_string starts, up to
   the point where  u_receive_result is called, and NULL
   after unity_parse_string finishes. */
static UnitExpression* parse_result = NULL;
void u_receive_result(float log_factor, const Unit* unit_sequence) 
{
    // It's a consequence of the grammar that u_receive_result will be
    // called precisely once per successful parse (though zero or one
    // times in an unsuccessful parse)
    assert(parse_result == NULL);
    if ((parse_result = (UnitExpression*)malloc(sizeof(UnitExpression))) == NULL) {
        fprintf(stderr, "unity: can't allocate %lu bytes for parse result!  Bailing...\n",
                (unsigned long)sizeof(UnitExpression));
        exit(1);
    }
    assert(parse_result != NULL);

    parse_result->log_factor = log_factor;
    parse_result->unit_sequence = unit_sequence;

    return;
}

/**
 * Parse a string.  If this returns with an error, then there should
 * be an error message available from {@link #unity_parse_error}.
 *
 * <p>The syntax is one of the codes obtained from {@link #unity_identify_parser}.
 * @param unit_string the string to be parsed
 * @param syntax the syntax to be used to parse the string
 * @return a parsed representation of the unit expression, or NULL on error
 */
const UnitExpression* unity_parse_string(const char* unit_string, const int syntax)
{
    parsef_ptr yyparse;
    const UnitExpression* rval;
    
    assert(parse_result == NULL);
    assert(TRACK_UNIT_EMPTY_P);

    if ((yyparse = get_parser_for_syntax(syntax)) != NULL) {
        lex_setup(unit_string);
        if ((*yyparse)() == 0 && !lex_unexpected_character()) {
            // successful parse
            rval = parse_result;
            // ensure that parse_result is clear, to make sure nothing else can get to it
            parse_result = NULL;
        } else {
            if (lex_unexpected_character()) {
                // this is a lexing error, not a parsing error
                char buf[64];           /* long enough? */
                snprintf(buf, 64, "Unexpected character '%c'", lex_unexpected_character());
                yyerror(buf);
            }
            /* parse error.  If there is a parse_result (that is,
             * u_receive_result was called, and what has happened is
             * that there was more in the parse string than was
             * matched by the 'begin' production), then free this and
             * the Unit objects it manages (which includes the tracked
             * units).  If there is not, then free each of the
             * (orphaned) Unit instances held in the unit_tracking
             * list.
             */
            if (parse_result == NULL) {
                track_unit_free();
            } else {
                unity_free_expression(parse_result);
                parse_result = NULL;
            }
            rval = NULL;
        }
        // in either case, stop tracking these units
        // (this simply causes the tracking to forget about these units, but mustn't free them)
        track_unit(NULL);
    } else {
        // we didn't even start to parse, so there are no leftover units to track
        char buf[64];
        snprintf(buf, 64, "Can't find parser for syntax: %d", syntax);
        yyerror(buf);
        rval = NULL;
    }

    assert(parse_result == NULL);
    assert(TRACK_UNIT_EMPTY_P);

    return rval;
}

/**
 * Examine the unit_string to see if there's an SI prefix.  If there
 * is, return the base unit, after the prefix, and store into *pfx the
 * power corresponding to the unit.  Return a pointer to the beginning
 * of the base unit.  This does not match if there is no unit
 * following the prefix.  If there is no SI prefix, return NULL.
 *
 * That is, this effectively matches the regexp "^(da|[yzafpnumcdhkMGTPEZY])(.+)".
 */
static const char* find_prefix_and_base(const char* unit_string, int* pfx)
{
    const char* prefixes = "yzafpnumcdhkMGTPEZY";
    const char* base_unit = NULL;

    if (strcmp(unit_string, "da") == 0) {
        // arcane special case -- this is a deci-year...
        *pfx = u_prefix_to_power('d');
        base_unit = "a";
    } else if (strncmp(unit_string, "da", 2) == 0) {
        *pfx = u_prefix_to_power(UNITY_PREFIX_DECA);
        base_unit = unit_string+2;
    } else {
        const char *p;
        for (p=prefixes; *p!='\0'; p++) {
            if (*p == *unit_string) {
                *pfx = u_prefix_to_power(*unit_string);
                base_unit = unit_string+1;
                break;
            }
        }
    }
    if (base_unit != NULL && *base_unit != '\0') {
        return base_unit;
    } else {
        return NULL;
    }
}

/**
 * Indicates whether the unit is being used in a way which satisfies
 * the indicated constraints.  Units can be
 * <ol>
 * <li>'known', which means the (prose) definition of the syntax
 * recognises this unit and has something to say about it -- all the
 * syntaxes 'know' the metre;</li>
 * <li>'deprecated', which means that the definition acknowledge the
 * existence of the unit but advises against it -- for example the
 * 'erg' is either unknown or deprecated in this sense (below, we take
 * 'recommended' to be a synonym for 'not deprecated'); or</li>
 * <li>'unknown', which means that the definition knows nothing about
 * this unit -- the 'furlong' tends not to be well-used in the
 * literature.</li>
 * </ol>
 *
 * The checks are indicated by the bitwise OR of the following flags:
 *
 * <p><code>UNITY_CHECK_RECOGNISED</code>: the unit is a 'known' unit, in the
 * sense that it is listed in the specification of the corresponding
 * syntax.
 *
 * <p><code>UNITY_CHECK_RECOMMENDED</code>: the unit is a 'known' unit, and is
 * additionally not deprecated.
 *
 * <p><code>UNITY_CHECK_CONSTRAINTS</code>: the unit is being used in conformance
 * with any other constraints placed on it.  Most typically, it is either
 * allowed to have such a prefix, or it has no SI prefix.  Possibly
 * surprisingly, an 'unknown' unit has no (known) constraints, so it
 * satisfies all of them.
 *
 * <p>The constant <code>UNITY_CHECK_ALL</code> performs all checks.
 *
 * @param u the Unit to be checked
 * @param syntax one of the constants returned by {@link #unity_identify_parser}
 * @param flags the checks to be performed
 * @return true (non-zero) if the checks pass
 */
int unity_check_unit(const Unit* u, int syntax, const int flags) 
{
    const UnitRep* rep;
    int is_recognised;
    int is_recommended;
    int satisfies_constraints;
    int result;

    if (flags == 0) {
        // trivial check
        return 1;               /* JUMP OUT */
    }

    if (u->type == function_application_type) {
        // for now, simply deem all functions to be acceptable
        // XXX: refine function checking
        return 1;               /* JUMP OUT */
    }

    rep = u_get_unit_representation(u, syntax);
    if (rep == NULL) {
        is_recognised = 0;
        is_recommended = 0;
        satisfies_constraints = 1; /* no constraints, so all satisfied */
    } else {
        is_recognised = 1;
        is_recommended = !rep->is_deprecated_p;
        satisfies_constraints = u->u.prefix_power == 0 || rep->si_prefixes_p;
    }
    
    result = 1;                 /* true */
    if (flags & UNITY_CHECK_RECOGNISED)
        result &= is_recognised;
    if (flags & UNITY_CHECK_RECOMMENDED)
        result &= is_recommended;
    if (flags & UNITY_CHECK_CONSTRAINTS)
        result &= satisfies_constraints;
    return result;
}

/**
 * Indicates whether the units in the expression are being used in a way which satisfies
 * the indicated constraints.  This checks each unit using 
 * {@link #unity_check_unit}.
 *
 * @param ue the UnitExpression to be checked
 * @param syntax one of the constants returned by {@link #unity_identify_parser}
 * or <code>..._OGIP</code> or <code>..._CDS</code>
 * @param flags the checks to be performed
 * @return true (non-zero) if the checks pass
 */
int unity_check_expression(const UnitExpression* ue, const int syntax, const int flags) 
{
    const Unit* u;
    for (u=ue->unit_sequence; u!=NULL; u=u->next) {
        if (! unity_check_unit(u, syntax, flags)) {
            return 0;           /* false */
        }
    }
    return 1;
}

/**
 * Return a version string for the library.
 * The form of the string is unspecified, but is intended to be printable
 * @see unity_version_number
 */
const char* unity_version_string(void)
{
    return PACKAGE_STRING;
}

/**
 * Return a number which indicates the version and release numbers of the library.
 * The number is formed by major_version * 1e6 + minor_version * 1e3 + release_number
 * @see unity_version_string
 */
int unity_version_number(void)
{
    return PACKAGE_VERSION_INTEGER;
}

