package uk.me.nxg.unity;

import java.util.List;

/**
 * A parser for unit strings.  The parser will parse a single string.
 * <p>There are a number of different parsers available:
 * <dl>
 * <dt>fits</dt>
 * <dd>Parses unit string according to the prescriptions in the FITS
 * specification, v3.0, section 4.3 (W.D. Pence et al., A&amp;A 524,
 * A42, 2010. 
 * <a href='http://dx.doi.org/10.1051/0004-6361/201015362'
 *    >doi:10.1051/0004-6361/201015362</a>).</dd>
 * <dt>ogip</dt>
 * <dd>The format defined in OGIP memo OGIP/93-001, 1993
 * (<a href='ftp://legacy.gsfc.nasa.gov/fits_info/fits_formats/docs/general/ogip_93_001/ogip_93_001.ps'
 *    >postscript via FTP</a>)</dd>
 * <dt>cds</dt>
 * <dd>A syntax based on the CDS document <em>Standards for Astronomical Catalogues,
 * Version 2.0</em>, 2000, specifically section 3.2.  
 * See <a href='http://cdsweb.u-strasbg.fr/doc/catstd-3.2.htx'
 *     ><code>http://cdsweb.u-strasbg.fr/doc/catstd-3.2.htx</code></a></dd>
 * <dt>vou</dt>
 * <dd>This is intended to be as nearly as possible in the intersection of the
 * various grammars above. It is a strict subset of the FITS and CDS grammars
 * (in the sense that any VOUnit unit string is a valid FITS and CDS string,
 * too), and it is almost a subset of the OGIP grammar, except that it uses the
 * dot for multiplication rather than star.
 * See <a href='http://ivoa.net/Documents/VOUnits/'
 *   >IVOA VOUnits Proposed Recommendation</a>.</dd>
 * </dl>
 * <p>See the {@link uk.me.nxg.unity} package documentation for fuller discussion of the grammars.
 */
public class UnitParser {

    /** Indicates the parser for VOUnit-format strings */
    public static final String VOUNITS = "vou".intern();
    /** Indicates the parser for FITS-format strings */
    public static final String FITS = "fits".intern();
    /** Indicates the parser for OGIP-format strings */
    public static final String OGIP = "ogip".intern();
    /** Indicates the parser for CDS-format strings */
    public static final String CDS  = "cds".intern();
    /** Indicates the formatter (not parser) for LaTeX/siunitx output */
    public static final String LATEX = "latex".intern();


    private final String inputString;
    private final Parser yyparser;

    private UnitExpr parseResult = null;

    private Yylex helperLexer;  // for debugging, only

    /**
     * Create a new unit parser.
     * The possible parser syntaxes are available from
     * {@link #getParserTypes()}, or can alternatively be specified using the
     * class's constant field values <code>FITS</code>,
     * <code>OGIP</code> or <code>CDS</code>.
     * @param parserType one of the known parser types
     * @param unitString the string to be parsed
     * @throws UnitParserException if no parser can be created (typically because the required syntax is unknown)
     */
    public UnitParser(String parserType, String unitString)
            throws UnitParserException {
        inputString = unitString;
        yyparser = SyntaxFactory.createParser(parserType, new java.io.StringReader(inputString));
        if (yyparser == null) {
            throw new UnitParserException("Unknown syntax: " + parserType);
        }
    }

    /**
     * Retrieve the parsed result
     * @throws UnitParserException if the parse fails
     */
    public UnitExpr getParsed()
            throws UnitParserException
    {
        if (helperLexer != null) {
            throw new IllegalStateException("Can't mix getLexeme and getParsed in UnitParser");
        }
        if (parseResult == null) {
            yyparser.parse();
            parseResult = yyparser.getParseResult();
        }
        return parseResult;
    }

    /**
     * Indicate the set of available parsers.
     * The returned syntaxes are all valid as arguments to {@link SyntaxFactory#createParser}
     * @return a list of parser names
     */
    public static java.util.Set<String> getParserTypes() {
        return SyntaxFactory.getParserTypes();
    }

    /**
     * Indicates whether the argument is the name of a known parser.
     * @param parser the string name of a parser (eg "fits")
     * @return true if the parser is one of the types returned by
     * {@link #getParserTypes}
     */
    public static boolean isKnownParser(String parser) {
        return SyntaxFactory.isKnownParser(parser);
    }        

    /**
     * Return successive lexemes from the input.
     * Primarily for debugging.
     * @return a Lexeme object, or null when the lexemes are exhausted
     */
    public UnitParser.Lexeme getLexeme() {
        if (helperLexer == null) {
            helperLexer = new Yylex(new java.io.StringReader(inputString), yyparser);
        }
        int token;
        try {
            if ((token = helperLexer.yylex()) == 0) {
                helperLexer = null;
                return null;
            } else {
                return new UnitParser.Lexeme(token, yyparser.yylval.toString());
            }
        } catch (java.io.IOException e) {
            System.err.println("IOException: " + e);
            helperLexer = null;
            return null;
        }
    }

    /**
     * The library main program, for exploring the library's functionality.
     * <p>Usage:
     * <pre>
     * UnitParser [-isyntax] [-osyntax] [-v] unit-expression
     * </pre>
     * <p>Parse and redisplay the string expression given as argument.
     * The <code>-i</code> and <code>-o</code> flags specify the
     * input and output syntaxes.
     * <pre>
     * UnitParser -S
     * </pre>
     * <p>List the set of available syntaxes
     */
    public static void main(String[] args) {
        String inputSyntax = UnitParser.FITS;
        String outputSyntax = UnitParser.FITS;
        boolean showSyntaxes = false;
        boolean showLexemes = false;
        boolean validateParse = false;
        String unitExpression = null;

        for (String arg : args) {
            if (arg.charAt(0) == '-') {
                // Yes, there are getopt packages for Java, but (a) I
                // don't want anything fancy, and (b) the one I tried
                // was GPL and, on reflection, I think I'd best
                // release this as BSD.
                if (arg.length() < 2) {
                    Usage();
                }
                switch (arg.charAt(1)) {
                  case 'i':
                    inputSyntax = arg.substring(2);
                    break;
                  case 'o':
                    outputSyntax = arg.substring(2);
                    break;
                  case 'S':
                    showSyntaxes = true;
                    break;
                  case 'x':
                    showLexemes = true;
                    break;
                  case 'v':
                    validateParse = true;
                    break;
                  default:
                    Usage();
                    break;
                }
            } else {
                if (unitExpression != null) {
                    Usage();
                } else {
                    unitExpression = arg;
                }
            }
        }
        if (! (showSyntaxes || unitExpression != null)) {
            Usage();
        }

        if (showSyntaxes) {
            for (String syntax : SyntaxFactory.getParserTypes()) {
                System.out.print(" " + syntax);
            }
            System.out.println();
        } else if (showLexemes) {
            try {
                UnitParser p = new UnitParser(inputSyntax, unitExpression);
                UnitParser.Lexeme l;
                while ((l = p.getLexeme()) != null) {
                    System.out.println(l);
                }
            } catch (UnitParserException e) {
                System.err.println("Can't lex expression \"" + unitExpression + "\": " + e);
            }
        } else {
            try {
                UnitParser p = new UnitParser(inputSyntax, unitExpression);
                UnitExpr e = p.getParsed();
                System.out.println(e.toString(outputSyntax));
                if (validateParse) {
                    System.out.println("check: all units recognised?        " 
                                       + (e.allUnitsRecognised(inputSyntax) ? "yes" : "no"));
                    System.out.println("check: all units recommended?       " 
                                       + (e.allUnitsRecommended(inputSyntax) ? "yes" : "no"));
                    System.out.println("check: all constraints satisfied?   " 
                                       + (e.allUsageConstraintsSatisfied(inputSyntax) ? "yes" : "no"));
                }
            } catch (UnitParserException e) {
                System.err.println("Error parsing units: " + e.getMessage());
            }
        }
    }
            
    private static void Usage() 
    {
        System.err.println("Usage:");
        System.err.println("    UnitParser [-isyntax] [-osyntax] [-v] unit-expression");
        System.err.println("    UnitParser -S");
        System.exit(1);
    }
        
    /**
     * A single lexeme.
     * This is primarily for debugging, and may be removed or hidden in future.
     */
    public static class Lexeme {
        public final int token;
        public final String value;
        private Lexeme(int token, String value) {
            this.token = token;
            this.value = value;
        }
        public String toString() {
            return Token.lookup(token);
        }
    }
}
